import pytz

from bolinette import core
from bolinette.decorators import service


@service('tz')
class TimezoneService(core.SimpleService):
    @staticmethod
    async def get_all():
        return [tz for tz in pytz.all_timezones]

    @staticmethod
    async def is_valid(key):
        return key in pytz.all_timezones



from bolinette import Console
from bolinette.decorators import command


@command('hello', 'Says hello!')
@command.argument('argument', 'name', summary='Who I am greeting?')
async def hello(name: str):
    console = Console()
    console.print(f'Hello {name}!')



import inspect as _inspect
from collections.abc import Callable, Awaitable
from typing import Literal, Union

from bolinette import blnt, core, web, BolinetteExtension
from bolinette.blnt.commands import Command as _Command, Argument as _Argument
from bolinette.utils import InitProxy as _InitProxy


def model(model_name: str, *,
          mixins: list[str] = None,
          database: str = 'default',
          model_type: Literal['relational', 'collection'] = 'relational',
          definitions: Literal['ignore', 'append', 'overwrite'] = 'ignore',
          join_table: bool = False):
    def decorator(model_cls: type['core.Model']):
        model_cls.__blnt__ = core.ModelMetadata(model_name, database, model_type == 'relational',
                                                join_table, mixins or [], definitions)
        blnt.cache.models[model_name] = model_cls
        return model_cls
    return decorator


def model_property(function):
    return core.ModelProperty(function.__name__, function)


class _MixinDecorator:
    def __call__(self, mixin_name: str):
        def decorator(mixin_cls: type['core.Mixin']):
            blnt.cache.mixins[mixin_name] = mixin_cls
            return mixin_cls
        return decorator

    @staticmethod
    def service_method(func: Callable):
        return core.MixinServiceMethod(func.__name__, func)


mixin = _MixinDecorator()


def init_func(*, extension: BolinetteExtension = None):
    def decorator(func: Callable[['blnt.BolinetteContext'], Awaitable[None]]):
        blnt.cache.init_funcs.append(blnt.InitFunc(func, extension))
        return func
    return decorator


def seeder(func):
    blnt.cache.seeders.append(func)
    return func


def service(service_name: str, *, model_name: str = None):
    def decorator(service_cls: type[Union['core.Service', 'core.SimpleService']]):
        service_cls.__blnt__ = core.ServiceMetadata(service_name, model_name or service_name)
        blnt.cache.services[service_name] = service_cls
        return service_cls
    return decorator


def controller(controller_name: str, path: str = None, *,
               namespace: str = '/api', use_service: bool = True,
               service_name: str = None, middlewares: Union[str, list[str]] = None):
    if path is None:
        path = f'/{controller_name}'
    if service_name is None:
        service_name = controller_name
    if middlewares is None:
        middlewares = []
    if isinstance(middlewares, str):
        middlewares = [middlewares]

    def decorator(controller_cls: type['web.Controller']):
        controller_cls.__blnt__ = web.ControllerMetadata(
            controller_name, path, use_service, service_name, namespace, middlewares)
        blnt.cache.controllers[controller_name] = controller_cls
        return controller_cls
    return decorator


def route(path: str, *, method: web.HttpMethod, expects: 'web.Expects' = None, returns: 'web.Returns' = None,
          middlewares: Union[str, list[str]] = None):
    if middlewares is None:
        middlewares = []
    if isinstance(middlewares, str):
        middlewares = [middlewares]

    def decorator(route_function: Callable):
        if not isinstance(route_function, _InitProxy) and not _inspect.iscoroutinefunction(route_function):
            raise ValueError(f'Route "{route_function.__name__}" must be an async function')
        if expects is not None and not isinstance(expects, web.Expects):
            raise ValueError(f'Route "{route_function.__name__}": expects argument must be of type web.Expects')
        if returns is not None and not isinstance(returns, web.Returns):
            raise ValueError(f'Route "{route_function.__name__}": expects argument must be of type web.Returns')
        inner_route = None
        if isinstance(route_function, _InitProxy):
            inner_route = route_function
        docstring = route_function.__doc__
        return _InitProxy(web.ControllerRoute, func=route_function, path=path, method=method,
                          docstring=docstring, expects=expects, returns=returns, inner_route=inner_route,
                          middlewares=middlewares)
    return decorator


def get(path: str, *, returns: 'web.Returns' = None,
        middlewares: Union[str, list[str]] = None):
    return route(path, method=web.HttpMethod.GET, expects=None, returns=returns, middlewares=middlewares)


def post(path: str, *, expects: 'web.Expects' = None, returns: 'web.Returns' = None,
         middlewares: Union[str, list[str]] = None):
    return route(path, method=web.HttpMethod.POST, expects=expects, returns=returns, middlewares=middlewares)


def put(path: str, *, expects: 'web.Expects' = None, returns: 'web.Returns' = None,
        middlewares: Union[str, list[str]] = None):
    return route(path, method=web.HttpMethod.PUT, expects=expects, returns=returns, middlewares=middlewares)


def patch(path: str, *, expects: 'web.Expects' = None, returns: 'web.Returns' = None,
          middlewares: Union[str, list[str]] = None):
    return route(path, method=web.HttpMethod.PATCH, expects=expects, returns=returns, middlewares=middlewares)


def delete(path: str, *, returns: 'web.Returns' = None,
           middlewares: Union[str, list[str]] = None):
    return route(path, method=web.HttpMethod.DELETE, expects=None, returns=returns, middlewares=middlewares)


def middleware(name: str, *, priority: int = 100, auto_load: bool = False, loadable: bool = True):
    def decorator(middleware_cls: type['web.Middleware']):
        middleware_cls.__blnt__ = web.MiddlewareMetadata(name, priority, auto_load, loadable)
        blnt.cache.middlewares[name] = middleware_cls
        return middleware_cls
    return decorator


def topic(topic_name: str):
    def decorator(topic_cls: type['web.Topic']):
        topic_cls.__blnt__ = web.TopicMetadata(topic_name)
        blnt.cache.topics[topic_name] = topic_cls
        return topic_cls
    return decorator


def channel(rule: str):
    def decorator(channel_function: Callable):
        return web.TopicChannel(channel_function, rule)
    return decorator


class _CommandDecorator:
    @staticmethod
    def _create_command(func):
        return _Command(func.__name__, func)

    def __call__(self, name: str, summary: str, *, run_init=False):
        def decorator(arg):
            if isinstance(arg, _Command):
                cmd = arg
            elif _inspect.isfunction(arg):
                cmd = self._create_command(arg)
            else:
                raise ValueError('@command must only decorate functions or async functions')
            cmd.path = name
            cmd.summary = summary
            cmd.run_init = run_init
            blnt.cache.commands[cmd.name] = cmd
            return cmd
        return decorator

    def argument(self, arg_type: Literal['argument', 'option', 'flag', 'count'],
                 name: str, *, flag: str = None, summary: str = None,
                 value_type: type = None, default=None, choices: list = None):
        def decorator(arg):
            if isinstance(arg, _Command):
                cmd = arg
            elif _inspect.isfunction(arg):
                cmd = self._create_command(arg)
            else:
                raise ValueError('@command.argument must only decorate function or async functions')
            if arg_type not in ['argument', 'option', 'flag', 'count']:
                raise ValueError(f'Command {cmd.name}: {arg_type} is not a valid argument type')
            cmd.args = [_Argument(arg_type, name, flag=flag, summary=summary, value_type=value_type,
                                         default=default, choices=choices)] + cmd.args
            return cmd
        return decorator


command = _CommandDecorator()



import random
import string

import pytest

from bolinette import Bolinette
from bolinette.testing import BolitestClient

bolinette_app = None


@pytest.fixture
def client(loop):
    global bolinette_app
    if bolinette_app is None:
        bolinette_app = Bolinette(profile='test', overrides={
            'secret_key': ''.join(random.choices(string.ascii_letters + string.digits, k=32))
        })
        bolinette_app.init_bolinette()

    async def create_client():
        return BolitestClient(bolinette_app, loop)
    bolinette_app.init_extensions()
    return loop.run_until_complete(create_client())



from bolinette.blnt.database.engines.engine import DatabaseEngine
from bolinette.blnt.database.engines.relational import RelationalDatabase
from bolinette.blnt.database.engines.collection import CollectionDatabase



import re
from typing import Any, Optional, Literal

import yaml
from aiohttp_swagger import setup_swagger

from bolinette import blnt, web, types, mapping
from bolinette.decorators import get
from bolinette.utils import paths, files


class Documentation:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context
        self.swagger_path = self.context.instance_path('swagger.yaml')
        self._path_param_regex = re.compile(r'{([^}]*)}')
        self._response_regex = re.compile(r'^-response ([\d]{3})(?: ([^:]*))?(?:: ?(.*))?$')
        self._response_type_regex = re.compile(r'file\[([^]]*)]')
        self._response_returns_regex = re.compile(r'returns')
        self._type_map = {
            types.db.Integer: {'type': 'integer'},
            types.db.Boolean: {'type': 'boolean'},
            types.db.String: {'type': 'string'},
            types.db.Email: {'type': 'string', 'format': 'email'},
            types.db.Float: {'type': 'number', 'format': 'float'},
            types.db.Date: {'type': 'string', 'format': 'date-time'},
            types.db.Password: {'type': 'string', 'format': 'password'}
        }

    def build(self):
        self.context.logger.info('Building API documentation')
        content = {
            'openapi': '3.0.0',
            'info': {
                'title': self.context.manifest.get('name', 'Bolinette App'),
                'description': self.context.manifest.get('desc', 'My web app built with the Bolinette framework'),
                'version': self.context.manifest.get('version', '0.0.1')
            },
            'servers': [{'url': f'http://localhost:{self.context.env.get("port", 5000)}'}],
            'paths': self._build_routes(),
            'components': {
                'schemas': self._build_schemas()
            }
        }
        files.write(self.swagger_path, yaml.safe_dump(content))

    def _build_routes(self):
        routes = {}
        for path, method, route in self.context.resources.routes:
            self._build_route(path, method, route, routes)
        return routes

    def _build_route(self, path: str, method: web.HttpMethod, route: web.ControllerRoute, routes: dict[str, Any]):
        if route.controller is not None:
            if not path:
                path = '/'
            if path not in routes:
                routes[path] = {}
            docs = {
                'tags': [f'{route.controller.__blnt__.name} controller']
            }
            parsed_docs = self._parse_docs(route.docstring, route)
            if len(parsed_docs) > 0:
                docs.update(parsed_docs)
            if ('responses' not in docs or len(docs['responses']) <= 0) and route.returns:
                if 'responses' not in docs:
                    docs['responses'] = {}
                ref = self._build_ref(route, 'response')
                if len(ref) > 0:
                    docs['responses'][200] = {'content': {'application/json': {'schema': ref}}}
            parameters = self._parse_path(path)
            if len(parameters) > 0:
                docs['parameters'] = parameters
            routes[path][method.name.lower()] = docs
        if route.inner_route is not None:
            self._build_route(path, method, route.inner_route, routes)

    def _parse_docs(self, docstring: Optional[str], route: web.ControllerRoute):
        if not docstring:
            return {}
        docs = {}
        parsed = [s.strip('\n ') for s in docstring.split('\n\n')]
        doc_index = 0
        for part in parsed:
            self._parse_doc_line(part, docs, doc_index, route)
            doc_index += 1
        return docs

    def _parse_doc_line(self, part: str, docs: dict[str, Any], index: int, route: web.ControllerRoute):
        if index == 0:
            docs['summary'] = part
            return
        if part.startswith('-'):
            lines = [line.strip() for line in part.split('\n')]
            commands = []
            for line in lines:
                if line.startswith('-'):
                    commands.append(line)
                else:
                    commands[-1] += f' {line}'
            for command in commands:
                if command.startswith('-response'):
                    self._parse_responses(command, docs, route)
            return
        if 'description' not in docs:
            docs['description'] = ''
        if len(docs['description']) > 0:
            docs['description'] += '\n\n'
        docs['description'] += part

    def _parse_responses(self, text: str, docs: dict[str, Any], route: web.ControllerRoute):
        if (match := self._response_regex.match(text)) is not None:
            code = match.group(1)
            res_type = match.group(2)
            text = match.group(3)
            if 'responses' not in docs:
                docs['responses'] = {}
            response = {}
            if text:
                response['description'] = text
            if res_type:
                if self._response_returns_regex.match(res_type) is not None:
                    ref = self._build_ref(route, 'response')
                    if len(ref) > 0:
                        response['content'] = {'application/json': {'schema': ref}}
                elif (match := self._response_type_regex.match(res_type)) is not None:
                    if mime := match.group(1):
                        response['content'] = {mime: {'schema': {'type': 'string'}}}
            if len(response) > 0:
                docs['responses'][code] = response

    @staticmethod
    def _build_ref(route: web.ControllerRoute, schema_type: Literal['response', 'payload']):
        returns = route.returns
        if returns:
            ref = {'$ref': f'#/components/schemas/{schema_type}.{returns.model}.{returns.key}'}
            if returns.as_list:
                return {'type': 'array', 'items': ref}
            return ref
        return {}

    def _parse_path(self, path: str):
        parameters = []
        for match in self._path_param_regex.finditer(path):
            param, *args = match.group(1).split(':')
            parameters.append({
                'name': param,
                'in': 'path',
                'required': True
            })
        return parameters

    def _build_schemas(self):
        schemas = {}
        collections = {
            'payloads': self.context.mapper.payloads,
            'response': self.context.mapper.responses
        }
        include_defs = {'payloads': False, 'response': True}
        include_fks = {'payloads': True, 'response': False}
        for def_type, collection in collections.items():
            inc_defs = include_defs[def_type]
            inc_fks = include_fks[def_type]
            for model, key, definition in collection:
                properties = {}
                for field in definition.fields:
                    if isinstance(field, mapping.Field):
                        properties[field.name] = self._type_map[field.type]
                    elif isinstance(field, mapping.Definition):
                        if inc_defs:
                            properties[field.name] = {
                                '$ref': f'#/components/schemas/{def_type}.{field.model_name}.{field.model_key}'
                            }
                        if inc_fks and isinstance(field, mapping.Reference):
                            properties[field.foreign_key] = {
                                'type': 'int'
                            }
                    elif isinstance(field, mapping.List) and inc_defs:
                        elem = field.element
                        if isinstance(elem, mapping.Definition):
                            properties[field.name] = {
                                'type': 'array',
                                'items': {
                                    '$ref': f'#/components/schemas/{def_type}.{elem.model_name}.{elem.model_key}'
                                }
                            }
                schema = {
                    'type': 'object',
                    'properties': properties
                }
                schemas[f'{def_type}.{model}.{key}'] = schema
        return schemas

    def setup(self):
        if paths.exists(self.swagger_path):
            setup_swagger(self.context.app, swagger_url='/api', ui_version=3, swagger_from_file=self.swagger_path)
        else:
            no_docs_ctrl = NoDocsController(self.context)
            no_docs_route: web.ControllerRoute = no_docs_ctrl.get_no_docs.instantiate(controller=no_docs_ctrl)
            no_docs_route.setup()


class NoDocsController(web.Controller):
    __blnt__ = web.ControllerMetadata('no_docs', '', False, '', '/api', [])

    def __init__(self, context: 'blnt.BolinetteContext'):
        super().__init__(context)

    @get('')
    async def get_no_docs(self):
        params = {
            'name': self.context.manifest.get('name', 'Bolinette App'),
            'desc': self.context.manifest.get('desc', 'My web app built with the Bolinette framework'),
            'version': self.context.manifest.get('version', '0.0.1')
        }
        return self.response.render_template('no_docs.html.jinja2', params,
                                             self.context.internal_files_path('templates'))



import traceback

import aiohttp_cors
from aiohttp import web as aio_web
from aiohttp.web_request import Request
from aiohttp.web_urldispatcher import Resource, ResourceRoute

from bolinette import console, blnt, web
from bolinette.exceptions import APIError, APIErrors, InternalError
from bolinette.utils.serializing import serialize


class BolinetteResources:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context
        self._routes: dict[str, dict[web.HttpMethod, web.ControllerRoute]] = {}
        self._aiohttp_resources: dict[str, 'BolinetteResource'] = {}
        self.cors = None

    @property
    def routes(self):
        for path, methods in self._routes.items():
            for method, route in methods.items():
                yield path, method, route

    def _setup_cors(self):
        parsed_conf = {}
        conf = self.context.env.get_all(startswith='cors.')
        for key, value in conf.items():
            keys = key.split('.', maxsplit=2)
            if len(keys) == 3:
                _, site, prop = keys
                if site not in parsed_conf:
                    parsed_conf[site] = {}
                parsed_conf[site][prop] = value
        cors = {}
        for site, conf in parsed_conf.items():
            cors[site] = aiohttp_cors.ResourceOptions(
                allow_credentials=conf.get('allow_credentials', False),
                expose_headers=conf.get('expose_headers', ()),
                allow_headers=conf.get('allow_headers', ()),
                max_age=conf.get('max_age', None),
                allow_methods=conf.get('allow_methods', None),
            )
        return cors

    def init_web(self, app: aio_web.Application):
        self._aiohttp_resources = {}
        self.cors = aiohttp_cors.setup(app, defaults=self._setup_cors())
        for path, methods in self._routes.items():
            for method, route in methods.items():
                if path not in self._aiohttp_resources:
                    self._aiohttp_resources[path] = BolinetteResource(
                        self.cors.add(self.context.app.router.add_resource(path)))
                handler = RouteHandler(route)
                self._aiohttp_resources[path].routes[method] = self.cors.add(
                    self._aiohttp_resources[path].resource.add_route(method.http_verb, handler.__call__))

    def add_route(self, path: str, route: 'web.ControllerRoute'):
        if path not in self._routes:
            self._routes[path] = {}
        self._routes[path][route.method] = route


class BolinetteResource:
    def __init__(self, resource: Resource):
        self.resource = resource
        self.routes: dict[web.HttpMethod, ResourceRoute] = {}


class RouteHandler:
    def __init__(self, route: 'web.ControllerRoute'):
        self.controller = route.controller
        self.route = route

    async def __call__(self, request: Request):
        context: blnt.BolinetteContext = request.app['blnt']
        params = {
            'match': {},
            'query': {},
            'request': request
        }
        for key in request.match_info:
            params['match'][key] = request.match_info[key]
        for key in request.query:
            params['query'][key] = request.query[key]
        try:
            resp = await self.route.call_middleware_chain(request, params)
            return resp
        except (APIError, APIErrors) as ex:
            res = web.Response(context).from_exception(ex)
            if context.env['debug']:
                stack = traceback.format_exc()
                if isinstance(ex, InternalError):
                    console.error(stack)
                res.content['trace'] = stack.split('\n')
            serialized, mime = serialize(res.content, 'application/json')
            web_response = aio_web.Response(text=serialized, status=res.code, content_type=mime)
            return web_response



from enum import auto, unique, Enum


@unique
class HttpMethod(Enum):
    GET = auto()
    POST = auto()
    PUT = auto()
    PATCH = auto()
    DELETE = auto()

    @property
    def http_verb(self):
        return self.name



from collections.abc import Iterator
from typing import Union, Literal, Optional

from bolinette import core, blnt
from bolinette.blnt.database.engines import DatabaseEngine

MappingPyTyping = list[Union['types.mapping.Column', 'types.mapping.Field',
                             'types.mapping.List', 'types.mapping.Definition']]
MappingListPyTyping = Union[MappingPyTyping, tuple[str, MappingPyTyping]]


class Model:
    __blnt__: 'ModelMetadata' = None

    def __init__(self, database: 'DatabaseEngine'):
        self.__props__ = ModelProps(self, database)
        self.__repo__: Optional[core.Repository] = None

    def payloads(self) -> MappingListPyTyping:
        pass

    def responses(self) -> MappingListPyTyping:
        pass

    def get_mixin(self, name: str):
        return self.__props__.mixins[name]

    def __repr__(self):
        return f'<Model {self.__blnt__.name}>'


class ModelMetadata:
    def __init__(self, name: str, database: str, relational: bool, join: bool, mixins: list[str],
                 merge_defs: Literal['ignore', 'append', 'overwrite']):
        self.name = name
        self.database = database
        self.relational = relational
        self.join = join
        self.mixins = mixins
        self.merge_defs = merge_defs


class ModelProps(blnt.Properties):
    def __init__(self, model: Model, database: 'DatabaseEngine'):
        super().__init__(model)
        self.model = model
        self.database = database
        self.mixins: dict[str, core.Mixin] = {}
        self.primary: Optional[list['core.models.Column']] = None
        self.entity_key: Optional[list['core.models.Column']] = None

    def get_columns(self) -> Iterator[tuple[str, 'core.models.Column']]:
        return self._get_attributes_of_type(self.parent, core.models.Column)

    def get_relationships(self) -> Iterator[tuple[str, 'core.models.Relationship']]:
        return self._get_attributes_of_type(self.parent, core.models.Relationship)

    def get_properties(self) -> Iterator[tuple[str, 'ModelProperty']]:
        return self._get_cls_attributes_of_type(type(self.parent), ModelProperty)

    def get_back_refs(self) -> Iterator[tuple[str, 'core.models.ColumnList']]:
        return self._get_attributes_of_type(self.parent, core.models.ColumnList)


class ModelProperty:
    def __init__(self, name, function):
        self.name = name
        self.function = function

    def __repr__(self):
        return f'<ModelProperty {self.name}>'



from functools import wraps
from collections.abc import Callable, Awaitable
from typing import Any

from bolinette.testing import BolitestClient
from bolinette.utils.functions import async_invoke


def bolitest(*, before: Callable[[Any], Awaitable[None]] = None,
             after: Callable[[Any], Awaitable[None]] = None):
    def wrapper(func):
        @wraps(func)
        async def inner(client: BolitestClient):
            try:
                async with client:
                    await client.context.db.close_transaction()
                    await client.context.db.drop_all()
                    await client.context.db.create_all()
                    if before is not None:
                        await async_invoke(before, context=client.context, mock=client.mock)
                    await client.context.db.close_transaction()
                    await func(client=client)
                    if after is not None:
                        await async_invoke(after, context=client.context, mock=client.mock)
                    await client.context.db.close_transaction()
                    await client.context.db.drop_all()
            except Exception as e:
                await client.context.db.rollback_transaction()
                await client.context.db.drop_all()
                raise e
        return inner
    return wrapper



from typing import Any, Union

from aiohttp import web as aio_web

from bolinette import blnt, core, web, mapping, BolinetteExtension, Extensions
from bolinette.exceptions import InternalError
from bolinette.blnt.database import DatabaseManager
from bolinette.utils import paths, files
from bolinette.docs import Documentation


class BolinetteContext:
    def __init__(self, origin: str, *, extensions: list[BolinetteExtension] = None,
                 profile: str = None, overrides: dict[str, Any] = None):
        self._ctx = {}
        self._extensions = []
        self._tables: dict[str, Any] = {}
        self._models: dict[str, 'core.Model'] = {}
        self._repos: dict[str, 'core.Repository'] = {}
        self._services: dict[str, 'core.Service'] = {}
        self._controllers: dict[str, 'web.Controller'] = {}

        for ext in extensions or []:
            self.use_extension(ext)

        self.app = None
        self.cwd = paths.cwd()
        self.origin = origin
        self.env = blnt.Environment(self, profile=profile, overrides=overrides)
        self.manifest = files.read_manifest(
            self.root_path(), params={'version': self.env.get('version', '0.0.0')}) or {}
        self.logger = blnt.Logger(self)
        self.db = DatabaseManager(self)
        self.mapper = mapping.Mapper(self)
        self.validator = blnt.Validator(self)
        self.jwt = blnt.JWT(self)
        self.resources = web.BolinetteResources(self)
        self.docs = Documentation(self)
        self.sockets = web.BolinetteSockets(self)

    def init_web(self, app: aio_web.Application):
        self.app = app
        self.resources.init_web(app)

    def init_sockets(self, app: aio_web.Application):
        self.app = app
        self.sockets.init_socket_handler()

    def __getitem__(self, key):
        return self._ctx[key]

    def __setitem__(self, key, value):
        self._ctx[key] = value

    def model(self, name) -> Any:
        return self._models[name]

    def clear_extensions(self):
        self._extensions = []

    def use_extension(self, ext: BolinetteExtension):
        if ext not in self._extensions:
            for sub_ext in ext.dependencies:
                self.use_extension(sub_ext)
            self._extensions.append(ext)

    def has_extension(self, ext: Union[tuple[BolinetteExtension, ...], BolinetteExtension]):
        if Extensions.ALL in self._extensions:
            return True
        if isinstance(ext, BolinetteExtension):
            return ext in self._extensions
        if isinstance(ext, tuple):
            return len(ext) > 0 and any(self.has_extension(b) for b in ext)
        raise ValueError('BolinetteContext.has_extensions only accepts BolinetteExtensions instances')

    def add_model(self, name, model: 'core.Model'):
        self._models[name] = model

    def internal_path(self, *path):
        return paths.join(self.origin, *path)

    def internal_files_path(self, *path):
        return paths.join(self.origin, '_files', *path)

    def root_path(self, *path):
        return paths.join(self.cwd, *path)

    def instance_path(self, *path):
        return self.root_path('instance', *path)

    def env_path(self, *path):
        return self.root_path('env', *path)

    def static_path(self, *path):
        return self.root_path('static', *path)

    def templates_path(self, *path):
        return self.root_path('templates', *path)

    @property
    def models(self):
        return ((name, self._models[name]) for name in self._models)

    def table(self, name) -> Any:
        return self._tables[name]

    def add_table(self, name, table):
        self._tables[name] = table

    @property
    def tables(self):
        return ((name, self._tables[name]) for name in self._tables)

    def repo(self, name: str) -> Any:
        return self._repos[name]

    def add_repo(self, name, repo: 'core.Repository'):
        self._repos[name] = repo

    @property
    def repos(self):
        return ((name, self._repos[name]) for name in self._repos)

    def service(self, name) -> Any:
        if name not in self._services:
            raise InternalError(f'global.service.not_found:{name}')
        return self._services[name]

    def add_service(self, name, service: 'core.Service'):
        self._services[name] = service

    @property
    def services(self):
        return ((name, self._services[name]) for name in self._services)

    def controller(self, name) -> Any:
        return self._controllers[name]

    def add_controller(self, name, controller: 'web.Controller'):
        self._controllers[name] = controller

    @property
    def controllers(self):
        return ((name, self._controllers[name]) for name in self._controllers)



from bolinette.web.method import HttpMethod
from bolinette.web.controller import Controller, ControllerMetadata, ControllerRoute, Expects, Returns
from bolinette.web.middleware import Middleware, MiddlewareMetadata, InternalMiddleware
from bolinette.web.topic import Topic, TopicMetadata, TopicChannel
from bolinette.web.resources import BolinetteResources, BolinetteResource
from bolinette.web.sockets import BolinetteSockets
from bolinette.web.response import APIResponse, Response, Cookie



import sys


class Console:
    def __init__(self, *, flush: bool = False, debug: bool = False):
        self._flush = flush
        self._debug = debug

    def print(self, *values: tuple[str, ...], sep=' ', end='\n'):
        print(*values, sep=sep, end=end, flush=self._flush)

    def error(self, *values: tuple[str, ...], sep=' ', end='\n'):
        print(*values, file=sys.stderr, sep=sep, end=end, flush=self._flush)

    def debug(self, *values: tuple[str, ...], sep=' ', end='\n'):
        if self._debug:
            self.print(*values, sep=sep, end=end)


console = Console()



from typing import Any

from dateutil import parser as date_parser

from bolinette import blnt, types, exceptions, mapping, core
from bolinette.exceptions import APIErrors, APIError
from bolinette.utils.functions import is_db_entity


class Validator:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context

    async def validate_payload(self, model: str, key: str, values, patch=False, attr_prefix: str = None):
        if attr_prefix is None:
            attr_prefix = ''
        api_errors = APIErrors()
        valid = {}
        definition = self.context.mapper.payload(model, key)
        for field in definition.fields:
            field_key = field.key
            field_name = field.name
            if patch and field_name not in values:
                continue
            if isinstance(field, mapping.Reference):
                try:
                    valid[field_name] = await self._validate_linked_entity(field, field_name, values,
                                                                           f'{attr_prefix}{field_name}.')
                    if is_db_entity(valid[field_name]):
                        valid[field.foreign_key] = getattr(valid[field_name], field.reference_key)
                except APIErrors as errors:
                    for error in errors:
                        error[0] = f'{attr_prefix}{field_name}.' + error[0]
                        api_errors.append(error)
                except APIError as error:
                    api_errors.append(error)
            elif field_key is not None and field_name is not None:
                try:
                    valid[field_key] = self._validate_field(field, field_name, values)
                except APIError as ex:
                    api_errors.append(ex)
        if api_errors:
            raise api_errors
        return valid

    @staticmethod
    def _validate_field(field: 'mapping.MappingObject', name: str, values: dict[str, Any]):
        if field.required and name not in values:
            raise exceptions.ParamMissingError(name)
        value = values.get(name, field.default)
        if not value and not field.nullable:
            raise exceptions.ParamNonNullableError(name)
        if value is not None and isinstance(field, mapping.Field):
            if field.type == types.db.Date:
                value = date_parser.parse(value)
        return value

    async def _validate_linked_entity(self, field: 'mapping.Reference', name: str,
                                      values: dict[str, Any], attr_prefix: str):
        model: core.Model = self.context.model(field.model_name)
        repo: core.Repository = self.context.repo(field.model_name)
        if name not in values:
            if field.required:
                raise exceptions.ParamMissingError(name)
            else:
                return None
        obj = values[name]
        if not isinstance(obj, dict):
            raise exceptions.BadParamFormatError(name, 'dict')
        field_def = self.context.mapper.payload(field.model_name, field.model_key)
        cols = [field_def[col.name] for col in model.__props__.entity_key]
        keys = dict((col.key, obj.get(col.name, None)) for col in cols)
        entity = await repo.query().filter_by(**keys).first()
        if entity is None:
            if not field.create_if_not_found:
                raise exceptions.EntityNotFoundError(field.model_name, ','.join(keys.keys()), ','.join(keys.values()))
            return await self.validate_payload(field.model_name, field.model_key, obj,
                                               patch=False, attr_prefix=attr_prefix)
        return entity



from collections.abc import Callable, Awaitable
from typing import Type

from bolinette import blnt, core, web, BolinetteExtension
from bolinette.blnt.commands import Command


class BolinetteCache:
    def __init__(self):
        self.models: dict[str, Type['core.Model']] = {}
        self.mixins: dict[str, Type['core.Mixin']] = {}
        self.services: dict[str, Type['core.Service']] = {}
        self.controllers: dict[str, Type['web.Controller']] = {}
        self.middlewares: dict[str, Type['web.Middleware']] = {}
        self.topics: dict[str, Type['web.Topic']] = {}
        self.init_funcs: list[InitFunc] = []
        self.commands: dict[str, Command] = {}
        self.seeders = []


cache = BolinetteCache()


class InitFunc:
    def __init__(self, func: Callable[['blnt.BolinetteContext'], Awaitable[None]], ext: BolinetteExtension):
        self._func = func
        self._ext = ext

    @property
    def extension(self):
        return self._ext

    def __call__(self, context: 'blnt.BolinetteContext'):
        return self._func(context)

    def __str__(self):
        return self._func.__name__



class BolinetteExtension:
    def __init__(self, name: str, dependencies: list['BolinetteExtension'] = None):
        self.name = name
        self.dependencies = dependencies or []

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return (isinstance(other, BolinetteExtension) and self.name == other.name
                or isinstance(other, str) and self.name == other)

    def __repr__(self):
        return f'<Extensions {self.name}>'


class Extensions:
    MODELS = BolinetteExtension('models')
    WEB = BolinetteExtension('web', [MODELS])
    SOCKETS = BolinetteExtension('sockets', [MODELS])
    ALL = BolinetteExtension('all')



import bolinette
from bolinette.decorators import command


@command('run server', 'Run the internal server', run_init=True)
@command.argument('option', 'host', flag='H', summary='The host the server will listen to')
@command.argument('option', 'port', flag='p', summary='The port the server will listen on', value_type=int)
def run_server(blnt: 'bolinette.Bolinette', host: str = None, port: int = None):
    blnt.start_server(host=host, port=port)



from abc import ABC
from typing import Union

from bolinette import core, mapping, types


class MappingObject(ABC):
    def __init__(self, *, key=None, name=None, default=None, required=False,
                 nullable=True, function=False, formatting=False):
        self.key = key
        self.name = name
        self.default = default
        self.required = required
        self.nullable = nullable
        self.function = function
        self.formatting = formatting
        if not self.name and self.key:
            self.name = self.key


class Field(MappingObject):
    def __init__(self, field_type: 'types.db.DataType', *, key=None, name=None, default=None, required=False,
                 nullable=True, function=None, formatting=None):
        super().__init__(key=key, name=name, default=default, required=required,
                         nullable=nullable, function=function, formatting=formatting)
        self.type = field_type

    def __repr__(self):
        return f'<MappingField {self.key}:{self.type} -> {self.name}>'


class Column(Field):
    def __init__(self, column: 'core.models.Column', *, name=None, default=None, required=False,
                 function=None, formatting=None):
        super().__init__(column.type, key=column.name, name=name, default=default, required=required,
                         nullable=column.nullable, function=function, formatting=formatting)


class List(MappingObject):
    def __init__(self, element: Union['types.db.DataType', 'mapping.Definition'], *,
                 key=None, name=None, default=None, required=False, function=None, formatting=None):
        super().__init__(key=key, name=name, default=default, required=required,
                         function=function, formatting=formatting)
        self.element = element

    def __repr__(self):
        return f'<MappingList {self.name}:[{repr(self.element)}]>'


class Definition(MappingObject):
    def __init__(self, model_name, model_key='default', *, key=None, name=None, default=None,
                 required=False, function=None, formatting=None):
        super().__init__(key=key, name=name, default=default, required=required,
                         function=function, formatting=formatting)
        self.fields: list[MappingObject] = []
        self.model_name = model_name
        self.model_key = model_key

    def __repr__(self):
        return f'<MappingModel {self.model_name}.{self.model_key}>'

    def __len__(self):
        return len(self.fields)

    def __getitem__(self, key: Union[int, str]):
        if isinstance(key, int):
            return self.fields[key]
        if isinstance(key, str):
            try:
                return next((f for f in self.fields if f.key == key))
            except StopIteration:
                raise KeyError(f'KeyError: no {key} field in {self.model_name}.{self.model_key} definition')
        raise ValueError('key must be int or str')


class Reference(Definition):
    def __init__(self, relationship: 'core.models.Relationship', model_key='default', *, name=None, default=None,
                 required=False, function=None, formatting=None, create_if_not_found=False):
        super().__init__(relationship.target_model_name, model_key, key=relationship.name, name=name, default=default,
                         required=required, function=function, formatting=formatting)
        self.foreign_key = relationship.foreign_key.name
        self.reference_model = relationship.foreign_key.reference.target_model_name
        self.reference_key = relationship.foreign_key.reference.target_column_name
        self.create_if_not_found = create_if_not_found



from typing import Any


class Pagination:
    def __init__(self, items: list[Any], page: int, per_page: int, total: int):
        self.items = items
        self.page = page
        self.per_page = per_page
        self.total = total

    def __repr__(self) -> str:
        return f'<Pagination {self.page}:{self.per_page} {len(self.items)}/{self.total}>'


class PaginationParams:
    def __init__(self, page: int, per_page: int):
        self.page = page
        self.per_page = per_page

    def __repr__(self):
        return f'<Pagination {self.page}:{self.per_page}>'


class OrderByParams:
    def __init__(self, column: str, ascending: bool):
        self.column = column
        self.ascending = ascending

    def __repr__(self):
        return f'<OrderBy {self.column}:{"asc" if self.ascending else "desc"}>'



import json
from asyncio.events import AbstractEventLoop

from aiohttp import test_utils

import bolinette
from bolinette.testing import Mock
from bolinette.utils.serializing import serialize


class BolitestClient:
    def __init__(self, blnt_app: 'bolinette.Bolinette', loop: AbstractEventLoop):
        server = test_utils.TestServer(blnt_app.app, loop=loop)
        self.client = test_utils.TestClient(server, loop=loop)
        self.context = blnt_app.context
        self.mock = Mock(blnt_app.context)
        self.cookies = {}

    async def __aenter__(self):
        await self.client.start_server()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.client.close()

    def __await__(self):
        return self.__aenter__().__await__()

    def payload(self, **payload):
        if 'csrf_access_token' in self.cookies:
            payload['headers'] = {}
            payload['headers']['X-CSRF-TOKEN'] = self.cookies['csrf_access_token']
        return payload

    def parse_cookies(self, headers):
        for key, header in headers.items():
            if key == 'Set-Cookie':
                split = header.split(';')[0].split('=')
                self.cookies[split[0]] = split[1]

    async def post(self, path: str, data: dict = None, *, prefix='/api') -> dict:
        if data is None:
            data = {}
        res = await self.client.post(f'{prefix}{path}', data=serialize(data, 'application/json')[0],
                                     headers={'Content-Type': 'application/json'})
        self.parse_cookies(res.headers)
        text = await res.text()
        return json.loads(text)

    async def put(self, path: str, data: dict = None, *, prefix='/api') -> dict:
        if data is None:
            data = {}
        res = await self.client.put(f'{prefix}{path}', data=serialize(data, 'application/json')[0],
                                    headers={'Content-Type': 'application/json'})
        self.parse_cookies(res.headers)
        text = await res.text()
        return json.loads(text)

    async def patch(self, path: str, data: dict = None, *, prefix='/api') -> dict:
        if data is None:
            data = {}
        res = await self.client.patch(f'{prefix}{path}', data=serialize(data, 'application/json')[0],
                                      headers={'Content-Type': 'application/json'})
        self.parse_cookies(res.headers)
        text = await res.text()
        return json.loads(text)

    async def get(self, path: str, *, prefix='/api') -> dict:
        res = await self.client.get(f'{prefix}{path}')
        text = await res.text()
        return json.loads(text)

    async def delete(self, path: str, *, prefix='/api') -> dict:
        res = await self.client.delete(f'{prefix}{path}')
        text = await res.text()
        return json.loads(text)



import asyncio
import inspect
import sys
from argparse import ArgumentParser

import bolinette
from bolinette import Console
from bolinette.exceptions import InitError
from bolinette.blnt.commands import Command, Argument
from bolinette.utils.functions import async_invoke, invoke


class Parser:
    def __init__(self, blnt: 'bolinette.Bolinette', commands: dict[str, Command]):
        self.blnt = blnt
        self.commands = commands
        self._factories = {
            'argument': self._create_argument,
            'option': self._create_option,
            'flag': self._create_flag
        }
        self._console = Console()
        self._sub_commands = {}

    def run(self):
        tree = self._parse_commands()
        parser = ArgumentParser(description='Bolinette Web Framework')
        sub_parsers = parser.add_subparsers()
        self._build_parsers(tree, sub_parsers, [])
        parsed = vars(parser.parse_args())
        if '__blnt_cmd__' in parsed:
            cmd = parsed.pop('__blnt_cmd__')
            self._run_command(cmd, parsed)
        elif '__blnt_path__' in parsed:
            self._console.error(self._sub_commands[parsed['__blnt_path__']].format_help())
            sys.exit(1)
        else:
            self._console.error(parser.format_help())
            sys.exit(1)

    def _run_command(self, cmd: str, parsed: dict):
        command = self.commands[cmd]
        func = command.func
        parsed['blnt'] = self.blnt
        parsed['context'] = self.blnt.context
        if command.run_init:
            self.blnt.init_bolinette()
        if inspect.iscoroutinefunction(func):
            loop = asyncio.get_event_loop()
            loop.run_until_complete(async_invoke(func, **parsed))
        else:
            invoke(func, **parsed)

    def _parse_commands(self):
        command_tree = {}
        for _, command in self.commands.items():
            cur_node = command_tree
            path = command.path.split(' ')
            for elem in path[:-1]:
                if elem not in cur_node:
                    cur_node[elem] = {}
                cur_node = cur_node[elem]
            elem = path[-1]
            if elem in cur_node:
                raise InitError(f'Conflict with "{command.name}" command')
            cur_node[elem] = command
        return command_tree

    def _build_parsers(self, command_tree: dict, sub_parsers, path: list[str]):
        for name, elem in command_tree.items():
            if isinstance(elem, Command):
                sub_parser = sub_parsers.add_parser(name, help=elem.summary)
                for arg in elem.args:
                    self._factories[arg.arg_type](arg, sub_parser)
                sub_parser.set_defaults(__blnt_cmd__=elem.name)
            else:
                sub_parser = sub_parsers.add_parser(name, help=self._build_help(elem, path + [name]))
                sub_parser.set_defaults(__blnt_path__=name)
                self._sub_commands[name] = sub_parser
                self._build_parsers(elem, sub_parser.add_subparsers(), path + [name])
    
    @staticmethod
    def _build_help(command_tree: dict, path: list[str]):
        commands = [f'"{" ".join(path + [x])}"' for x in command_tree]
        return 'Sub-commands: ' + ', '.join(commands)

    @staticmethod
    def _create_parser_arg(arg: Argument, *, optional: bool = False, use_flag: bool = False,
                           action: str = None):
        args = []
        kwargs = {}
        if optional:
            args.append(f'--{arg.name}')
        else:
            args.append(arg.name)
        if use_flag and arg.flag is not None:
            args.append(f'-{arg.flag}')
        if arg.summary is not None:
            kwargs['help'] = arg.summary
        if action is not None:
            kwargs['action'] = action
        if arg.value_type is not None:
            kwargs['type'] = arg.value_type
        return args, kwargs

    @staticmethod
    def _create_argument(arg: Argument, parser: ArgumentParser):
        args, kwargs = Parser._create_parser_arg(arg, optional=False, use_flag=False, action=None)
        parser.add_argument(*args, **kwargs)

    @staticmethod
    def _create_option(arg: Argument, parser: ArgumentParser):
        args, kwargs = Parser._create_parser_arg(arg, optional=True, use_flag=True, action=None)
        parser.add_argument(*args, **kwargs)

    @staticmethod
    def _create_flag(arg: Argument, parser: ArgumentParser):
        args, kwargs = Parser._create_parser_arg(arg, optional=True, use_flag=True, action='store_true')
        parser.add_argument(*args, **kwargs)



import bolinette.commands.run_server
import bolinette.commands.init_db
import bolinette.commands.init_docs
import bolinette.commands.user_management
import bolinette.commands.create_files
import bolinette.commands.hello



from aiohttp import web as aio_web

from bolinette import web
from bolinette.decorators import topic


@topic('role')
class RoleTopic(web.Topic):
    async def validate_subscription(self, socket: aio_web.WebSocketResponse, payload, current_user) -> bool:
        if current_user is None:
            await self.send_error(socket, f'role.{payload["channel"]}.require_auth')
            return False
        if current_user.username != payload['channel']:
            await self.send_error(socket, f'role.{payload["channel"]}.forbidden')
            return False
        return True



from typing import Any

import jinja2
import yaml

from bolinette.utils import paths


def write(path, content, mode='w+'):
    with open(path, mode=mode) as file:
        file.write(content)


def append(path, content):
    write(path, content, mode='a+')


def read_file(path):
    try:
        with open(path) as f:
            return f.read()
    except FileNotFoundError:
        return None


def read_requirements(path):
    try:
        with open(paths.join(path, 'requirements.txt')) as f:
            return list(filter(lambda r: len(r), f.read().split('\n')))
    except FileNotFoundError:
        return []


def read_manifest(path, *, params: dict[str, Any] = None):
    try:
        with open(paths.join(path, 'manifest.blnt.yaml')) as f:
            raw = f.read()
            if params is not None:
                raw = render_string(raw, params)
            return yaml.safe_load(raw)
    except FileNotFoundError:
        return None


def render_template(workdir: str, path: str, params: dict[str, Any]):
    jinja_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(searchpath=workdir),
        keep_trailing_newline=True,
        trim_blocks=True,
        lstrip_blocks=True
    )
    return jinja_env.get_template(path).render(**params)


def render_string(string: str, params: dict[str, Any]):
    jinja_env = jinja2.Environment(
        loader=jinja2.BaseLoader,
        keep_trailing_newline=True,
        trim_blocks=True,
        lstrip_blocks=True
    )
    return jinja_env.from_string(string).render(**params)



import re
from collections.abc import Generator, Callable
from typing import Any, Optional, Union

from aiohttp.web_request import Request

from bolinette import blnt, core, web
from bolinette.exceptions import InitError
from bolinette.utils import functions


class Controller:
    __blnt__: 'ControllerMetadata' = None

    def __init__(self, context: 'blnt.BolinetteContext'):
        self.__props__ = ControllerProps(self)
        self.context = context
        self.response = web.Response(context)
        if self.__blnt__.use_service:
            self.defaults = ControllerDefaults(self)

    @property
    def service(self) -> 'core.Service':
        return self.context.service(self.__blnt__.service_name)

    def default_routes(self) -> list['web.ControllerRoute']:
        return []

    def __repr__(self):
        return f'<Controller {self.__blnt__.name} {self.__blnt__.path}>'


class ControllerMetadata:
    def __init__(self, name: str, path: str, use_service: bool, service_name: str,
                 namespace: str, middlewares: list[str]):
        self.name = name
        self.path = path
        self.use_service = use_service
        self.service_name = service_name
        self.namespace = namespace
        self.middlewares = middlewares


class ControllerProps(blnt.Properties):
    def __init__(self, controller):
        super().__init__(controller)

    def get_routes(self) -> Generator[tuple[str, 'ControllerRoute'], Any, None]:
        return self._get_attributes_of_type(self.parent, ControllerRoute)


class ControllerRoute:
    def __init__(self, controller: 'web.Controller', func: Callable, path: str, method: web.HttpMethod,
                 docstring: Optional[str], expects: 'Expects' = None, returns: 'Returns' = None,
                 inner_route: 'ControllerRoute' = None, middlewares: list[str] = None):
        self.controller = controller
        self.func = func
        self.path = path
        self.method = method
        self.docstring = docstring
        self.expects = expects
        self.returns = returns
        self.inner_route = inner_route
        self._mdw_defs = middlewares or []
        self.middlewares: list['web.Middleware'] = []

    def __repr__(self):
        return f'<Route {self.method.name} "{self.full_path}" {self._mdw_defs}>'

    @property
    def full_path(self):
        return f'{self.controller.__blnt__.namespace}{self.controller.__blnt__.path}{self.path}'

    def setup(self):
        self._init_middlewares(self.controller.context, self.controller.__blnt__.middlewares,
                               self._init_sys_middleware())
        self.controller.context.resources.add_route(self.full_path, self)
        if self.inner_route is not None:
            self.inner_route.setup()

    def _init_middlewares(self, context: 'blnt.BolinetteContext', from_controller: list[str], system: list[str]):
        for mdw in system:
            self._parse_middleware_options(mdw, context, True)
        for mdw in from_controller:
            self._parse_middleware_options(mdw, context)
        for mdw in self._mdw_defs:
            self._parse_middleware_options(mdw, context)
        self.middlewares = sorted(sorted(self.middlewares, key=lambda m: m.__blnt__.priority),
                                  key=lambda m: m.system_priority)

    def _init_sys_middleware(self):
        sys_mdw = [m.__blnt__.name for m in blnt.cache.middlewares.values() if m.__blnt__.auto_load]
        if self.expects is not None:
            model = self.expects.model
            key = self.expects.key if self.expects.key is not None else 'default'
            cmd = f'blnt_payload|model={model}|key={key}'
            if self.expects.patch:
                cmd += '|patch'
            sys_mdw.append(cmd)
        else:
            sys_mdw.append('blnt_payload')
        if self.returns is not None:
            model = self.returns.model
            key = self.returns.key if self.returns.key is not None else 'default'
            cmd = f'blnt_response|model={model}|key={key}'
            if self.returns.as_list:
                cmd += '|as_list'
            if self.returns.skip_none:
                cmd += '|skip_none'
            sys_mdw.append(cmd)
        else:
            sys_mdw.append('blnt_response')
        return sys_mdw

    def _parse_middleware_options(self, mdw: str, context, system: bool = False):
        name, *args = mdw.split('|')
        if name.startswith('!'):
            self.middlewares = list(filter(lambda m: m.__blnt__.name != name[1:], self.middlewares))
            return
        find = list(filter(lambda m: m.__blnt__.name == name, self.middlewares))
        if len(find) > 0:
            middleware = find[0]
            middleware.options = {}
        else:
            middleware = blnt.cache.middlewares[name](context)

        if not middleware.__blnt__.loadable and not system:
            raise InitError(f'[{type(self.controller).__name__}] Middleware '
                            f'"{mdw}" cannot be loaded in a controller')

        parsed_args = {}
        for arg in args:
            arg_n, *arg_v = arg.split('=', maxsplit=1)
            parsed_args[arg_n] = arg_v[0] if len(arg_v) else True
        def_options = middleware.define_options()
        for opt_name, option in def_options.items():
            if opt_name in parsed_args:
                middleware.options[opt_name] = option.validate(parsed_args[opt_name])
            elif option.required:
                raise InitError(f'[{type(self.controller).__name__}] Middleware "{mdw}" '
                                f'option "{opt_name}" is missing from declaration string')
            else:
                middleware.options[opt_name] = option.default
        self.middlewares.append(middleware)

    async def call_middleware_chain(self, request: Request, params: dict[str, Any]):
        handles = ([MiddlewareHandle(name=m.__blnt__.name, func=m.handle) for m in self.middlewares]
                   + [MiddlewareHandle('__ctrl_call', self._final_middleware_call)])
        return await self._call_middleware(handles, 0, request, params)

    async def _call_middleware(self, handles: list['MiddlewareHandle'], index: int, request: Request,
                               params: dict[str, Any]):
        async def _next(_request, _params):
            return await self._call_middleware(handles, index + 1, _request, _params)
        return await handles[index].func(request, params, _next)

    async def _final_middleware_call(self, _1, params: dict[str, Any], _2):
        return await functions.async_invoke(self.func, self.controller, **params)


class MiddlewareHandle:
    def __init__(self, name: str, func: Callable):
        self.name = name
        self.func = func


class Expects:
    def __init__(self, model: str, key: str = 'default', *, patch: bool = False):
        self.model = model
        self.key = key
        self.patch = patch


class Returns:
    def __init__(self, model: str, key: str = 'default', *, as_list: bool = False, skip_none: bool = False):
        self.model = model
        self.key = key
        self.as_list = as_list
        self.skip_none = skip_none


class ControllerDefaults:
    PARAM_NAME_DOT_REGEX = re.compile(r'\.')
    PARAM_NAME_CHAR_REGEX = re.compile(r'[^a-zA-Z0-9_]')

    def __init__(self, controller: Controller):
        self.controller = controller
        self.service: core.Service = controller.service

    def _get_url_keys(self, key: Optional[Union[str, list[str]]], *, route: str) -> list[str]:
        if key is None:
            entity_key = self.service.repo.model.__props__.entity_key
            if entity_key is None:
                raise InitError(f'Default route "{route}" for controller "{self.service.repo.model.__blnt__.name}" '
                                f'must define a key or set entity_key=True in one or more columns in the model')
            key = [m.name for m in entity_key]
        if not isinstance(key, list):
            key = [key]
        return key

    def _get_url(self, keys: list[str], *, prefix='') -> str:
        return f'{prefix}/' + '/'.join([f'{{{self._translate_param_name(k)}}}' for k in keys])

    def _get_match_params(self, keys: list[str], match: dict[str, Any]) -> dict[str, Any]:
        return dict((keys[i], match.get(self._translate_param_name(keys[i]))) for i in range(len(keys)))

    def _translate_param_name(self, param: str):
        return self.PARAM_NAME_CHAR_REGEX.sub('', self.PARAM_NAME_DOT_REGEX.sub('_', param))

    def get_all(self, returns='default', *, prefix='',
                middlewares: list[str] = None, docstring: str = None):
        model_name = self.service.__blnt__.model_name

        async def route(controller: web.Controller, **kwargs):
            resp = await functions.async_invoke(controller.service.get_all, **kwargs)
            return controller.response.ok(data=resp)

        docstring = docstring or f"""
        Gets all records from {model_name} collection

        -response 200 returns: The list of {model_name} entities
        """
        return ControllerRoute(self.controller, route, f'{prefix}', web.HttpMethod.GET, docstring,
                               returns=Returns(model_name, returns, as_list=True),
                               middlewares=middlewares)

    def get_one(self, returns='default', *, key: Union[str, list[str], None] = None,
                prefix='', middlewares: list[str] = None, docstring: str = None):
        model_name = self.service.__blnt__.model_name
        url_keys = self._get_url_keys(key, route='get_one')
        url = self._get_url(url_keys, prefix=prefix)

        async def route(controller: web.Controller, *, match, **kwargs):
            keys = self._get_match_params(url_keys, match)
            resp = await functions.async_invoke(controller.service.get_first_by_keys, keys, **kwargs)
            return controller.response.ok(data=resp)

        docstring = docstring or f"""
        Gets one record from {model_name} collection, identified by {key} field

        -response 200 returns: The {model_name} entity
        -response 404: The {model_name} identified by "{key}" was not found
        """
        return ControllerRoute(self.controller, route, url, web.HttpMethod.GET, docstring,
                               returns=Returns(model_name, returns),
                               middlewares=middlewares)

    def create(self, returns='default', expects='default', *, prefix='',
               middlewares: list[str] = None, docstring: str = None):
        model_name = self.service.__blnt__.model_name

        async def route(controller: web.Controller, payload, **kwargs):
            resp = await functions.async_invoke(controller.service.create, payload, **kwargs)
            return controller.response.created(messages=f'{controller.service.__blnt__.model_name}.created', data=resp)

        docstring = docstring or f"""
        Inserts a new entity in {model_name} collection

        -response 201 returns: The created {model_name} entity
        """
        return ControllerRoute(self.controller, route, f'{prefix}', web.HttpMethod.POST, docstring,
                               expects=Expects(self.service.__blnt__.model_name, expects),
                               returns=Returns(model_name, returns),
                               middlewares=middlewares)

    def update(self, returns='default', expects='default', *, key: str = None, prefix='',
               middlewares: list[str] = None, docstring: str = None):
        model_name = self.service.__blnt__.model_name
        url_keys = self._get_url_keys(key, route='update')
        url = self._get_url(url_keys, prefix=prefix)

        async def route(controller: web.Controller, payload, match, **kwargs):
            keys = self._get_match_params(url_keys, match)
            entity = await controller.service.get_first_by_keys(keys)
            resp = await functions.async_invoke(controller.service.update, entity, payload, **kwargs)
            return controller.response.ok(messages=f'{controller.service.__blnt__.model_name}.updated', data=resp)

        docstring = docstring or f"""
        Updates all fields from one {model_name} entity, identified by {key} field

        -response 200 returns: The updated {model_name} entity
        """
        return ControllerRoute(self.controller, route, url, web.HttpMethod.PUT, docstring,
                               expects=Expects(self.service.__blnt__.model_name, expects),
                               returns=Returns(model_name, returns),
                               middlewares=middlewares)

    def patch(self, returns='default', expects='default', *, key: str = None, prefix='',
              middlewares: list[str] = None, docstring: str = None):
        model_name = self.service.__blnt__.model_name
        url_keys = self._get_url_keys(key, route='patch')
        url = self._get_url(url_keys, prefix=prefix)

        async def route(controller: web.Controller, payload, match, **kwargs):
            keys = self._get_match_params(url_keys, match)
            entity = await controller.service.get_first_by_keys(keys)
            resp = await functions.async_invoke(controller.service.patch, entity, payload, **kwargs)
            return controller.response.ok(messages=f'{controller.service.__blnt__.model_name}.updated', data=resp)

        docstring = docstring or f"""
        Updates some fields from one {model_name} entity, identified by {key} field

        -response 200 returns: The updated {model_name} entity
        """
        return ControllerRoute(self.controller, route, url, web.HttpMethod.PATCH, docstring,
                               expects=Expects(self.service.__blnt__.model_name, expects, patch=True),
                               returns=Returns(model_name, returns),
                               middlewares=middlewares)

    def delete(self, returns='default', *, key: str = None, prefix='',
               middlewares: list[str] = None, docstring: str = None):
        model_name = self.service.__blnt__.model_name
        url_keys = self._get_url_keys(key, route='delete')
        url = self._get_url(url_keys, prefix=prefix)

        async def route(controller: web.Controller, match, **kwargs):
            keys = self._get_match_params(url_keys, match)
            entity = await controller.service.get_first_by_keys(keys)
            resp = await functions.async_invoke(controller.service.delete, entity, **kwargs)
            return controller.response.ok(messages=f'{controller.service.__blnt__.model_name}.deleted', data=resp)

        docstring = docstring or f"""
        Deletes on record from {model_name} collection, identified by {key} field

        -response 200 returns: The deleted {model_name} entity
        """
        return ControllerRoute(self.controller, route, url, web.HttpMethod.DELETE, docstring,
                               returns=Returns(model_name, returns),
                               middlewares=middlewares)



from collections.abc import Callable

from bolinette.blnt.commands import Argument


class Command:
    def __init__(self, name: str, func: Callable, path: str = None,
                 summary: str = None, args: list[Argument] = None, run_init: bool = False):
        self.name = name
        self.func = func
        self.path = path
        self.summary = summary
        self.args = args or []
        self.run_init = run_init



import sqlalchemy as _sqlalchemy


class DataType:
    def __init__(self, name, sqlalchemy_type):
        self.name = name
        self.sqlalchemy_type = sqlalchemy_type

    def of_type(self, cls):
        return type(self) is cls

    def __repr__(self):
        return self.name


String = DataType('String', _sqlalchemy.String)
Integer = DataType('Integer', _sqlalchemy.Integer)
Float = DataType('Float', _sqlalchemy.Float)
Boolean = DataType('Boolean', _sqlalchemy.Boolean)
Date = DataType('Date', _sqlalchemy.DateTime)
Email = DataType('Email', _sqlalchemy.String)
Password = DataType('Password', _sqlalchemy.String)



from bolinette import console
import getpass

import bolinette
from bolinette import Console, blnt
from bolinette import exceptions
from bolinette.decorators import command
from bolinette.exceptions import ParamConflictError, EntityNotFoundError, APIError, APIErrors


@command('user new', 'Add a user to the database', run_init=True)
@command.argument('argument', 'username', summary='The new user\'s username')
@command.argument('argument', 'email', summary='The new user\'s email')
@command.argument('option', 'roles', flag='r', summary='The user\'s roles, comma separated')
async def create_user(blnt: 'bolinette.Bolinette', username: str, email: str, roles: str = None):
    console = Console()
    user_service = blnt.context.service('user')
    role_service = blnt.context.service('role')
    while True:
        password = getpass.getpass('Choose password: ')
        password2 = getpass.getpass('Confirm password: ')
        if password == password2:
            break
        console.error('Passwords don\'t match')
    async with blnt.Transaction(blnt.context, print_error=False, propagate_error=False):
        user_roles = []
        if roles is not None:
            for role_name in [r.strip() for r in roles.split(',')]:
                try:
                    user_roles.append(await role_service.get_by_name(role_name))
                except EntityNotFoundError:
                    console.error(f'Role "{role_name}" does not exist')
                    exit(1)
        try:
            user = await user_service.create({
                'username': username,
                'password': password,
                'email': email
            })
            for role in user_roles:
                user.roles.append(role)
        except (APIError, APIErrors) as ex:
            if isinstance(ex, APIError):
                errors = [ex]
            else:
                errors = ex.errors
            for err in [err for err in errors if isinstance(err, ParamConflictError)]:
                console.error(f'Conflict: {err.message.split(":")[1]} already exists')
            exit(1)


@command('user list', 'Lists all usernames in database', run_init=True)
@command.argument('flag', 'roles', flag='r', summary='Prints user roles')
async def list_users(context: 'blnt.BolinetteContext', roles: bool):
    console = Console()
    user_service = context.service('user')
    for user in await user_service.get_all():
        roles_str = ''
        str_sep = ''
        if roles:
            roles_str = ','.join([r.name for r in user.roles])
            str_sep=':'
        console.print(user.username, roles_str, sep=str_sep)


@command('role list', 'Lists all usernames in database', run_init=True)
@command.argument('flag', 'users', flag='u', summary='Prints users assigned to each role')
async def list_roles(context: 'blnt.BolinetteContext', users: bool):
    console = Console()
    role_service = context.service('role')
    for role in await role_service.get_all():
        users_str = ''
        str_sep = ''
        if users:
            users_str = ','.join([r.username for r in role.users])
            str_sep=':'
        console.print(role.name, users_str, sep=str_sep)


@command('role assign', 'Assigns a user to a role', run_init=True)
@command.argument('argument', 'user', summary='The user\'s username')
@command.argument('argument', 'role', summary='The role\'s name')
@command.argument('flag', 'create', flag='c', summary='Create the role if it does not exist')
async def add_role(context: 'blnt.BolinetteContext', user: str, role: str, create: bool):
    console = Console()
    user_service = context.service('user')
    role_service = context.service('role')
    async with blnt.Transaction(context, print_error=False, propagate_error=False):
        try:
            user_ent = await user_service.get_by_username(user)
        except:
            console.error(f'User {user} does not exist')
            exit(1)
        role_ent = await role_service.get_by_name(role, safe=True)
        if role_ent is None:
            if create:
                role_ent = await role_service.create({'name': role})
                console.print(f'Created a new role: {role}')
            else:
                console.error(f'Role {role} does not exist')
                exit(1)
        if role_ent in user_ent.roles:
            console.print(f'User {user} is already assigned to role {role}')
            exit(1)
        user_ent.roles.append(role_ent)
    console.print(f'User {user} has been assigned to role {role}')


@command('role revoke', 'Rovokes a user from a role', run_init=True)
@command.argument('argument', 'user', summary='The user\'s username')
@command.argument('argument', 'role', summary='The role\'s name')
@command.argument('flag', 'prune', flag='p', summary='Deletes the role if noone is assigned to it')
async def revoke_role(context: 'blnt.BolinetteContext', user: str, role: str, prune: bool):
    console = Console()
    user_service = context.service('user')
    role_service = context.service('role')
    async with blnt.Transaction(context, print_error=False, propagate_error=False):
        try:
            user_ent = await user_service.get_by_username(user)
        except:
            console.error(f'User {user} does not exist')
            exit(1)
        try:
            role_ent = await role_service.get_first_by('name', role)
        except:
            console.error(f'Role {role} does not exist')
            exit(1)
        if role_ent not in user_ent.roles:
            console.error(f'User {user} is not assigned to role {role}')
            exit(1)
        user_ent.roles.remove(role_ent)
        console.print(f'User {user} has been rovoked from role {role}')
        if prune and len(role_ent.users) == 0:
            await role_service.delete(role_ent)
            console.print(f'Role {role} has been deleted')



from collections.abc import Callable, Awaitable
from typing import Any

from aiohttp import web as aio_web
from aiohttp.web_request import Request
from aiohttp.web_response import Response

from bolinette import web, blnt
from bolinette.blnt.objects import Pagination, PaginationParams, OrderByParams
from bolinette.decorators import middleware
from bolinette.exceptions import BadRequestError
from bolinette.utils.serializing import deserialize, serialize


@middleware('blnt_payload', priority=60, auto_load=False, loadable=False)
class PayloadMiddleware(web.InternalMiddleware):
    def define_options(self):
        return {
            'model': self.params.string(),
            'key': self.params.string(),
            'patch': self.params.bool()
        }

    async def handle(self, request, params, next_func):
        try:
            payload = await deserialize(request)
        except Exception:
            raise BadRequestError('global.payload.unserializable')
        if self.options['model'] is not None and self.options['key'] is not None:
            payload = await self.context.validator.validate_payload(self.options['model'], self.options['key'],
                                                                    payload, self.options['patch'])
        params['payload'] = payload
        return await next_func(request, params)


@middleware('blnt_response', priority=40, auto_load=False, loadable=False)
class ResponseMiddleware(web.InternalMiddleware):
    def define_options(self):
        return {
            'model': self.params.string(),
            'key': self.params.string(),
            'as_list': self.params.bool(),
            'skip_none': self.params.bool()
        }

    async def handle(self, request, params, next_func):
        async with blnt.Transaction(self.context):
            resp = await next_func(request, params)
        if resp is None:
            return aio_web.Response(status=204)
        elif isinstance(resp, aio_web.Response):
            return resp
        elif isinstance(resp, str):
            return aio_web.Response(text=resp, status=200, content_type='text/plain')
        elif not isinstance(resp, web.APIResponse):
            if self.options['model'] is not None:
                resp = web.Response(self.context).ok(data=resp)
            else:
                return aio_web.Response(text='global.response.unserializable', status=500, content_type='text/plain')

        content = resp.content

        if content.get('data') is not None and isinstance(content['data'], Pagination):
            content['pagination'] = {
                'page': content['data'].page,
                'per_page': content['data'].per_page,
                'total': content['data'].total
            }
            content['data'] = content['data'].items

        if self.options['model'] is not None:
            ret_def = self.context.mapper.response(self.options['model'], self.options['key'])
            if content.get('data') is not None:
                content['data'] = self.context.mapper.marshall(ret_def, content['data'],
                                                               skip_none=self.options.get('skip_none', False),
                                                               as_list=self.options.get('as_list', False))

        serialized, mime = serialize(content, 'application/json')

        web_response = aio_web.Response(text=serialized, status=resp.code, content_type=mime)
        for cookie in resp.cookies:
            if not cookie.delete:
                web_response.set_cookie(cookie.name, cookie.value,
                                        expires=cookie.expires.strftime('%a, %d %b %Y %H:%M:%S GMT'),
                                        path=cookie.path, httponly=cookie.http_only)
            else:
                web_response.del_cookie(cookie.name, path=cookie.path)

        return web_response


@middleware('blnt_headers', priority=10, auto_load=True, loadable=False)
class HeadersMiddleware(web.InternalMiddleware):
    async def handle(self, request: Request, params: dict[str, Any],
                     next_func: Callable[[Request, dict[str, Any]], Awaitable[Response]]):
        params['headers'] = {}
        for key in request.headers:
            params['headers'][key] = request.headers[key]
        return await next_func(request, params)


@middleware('blnt_query_pagination', priority=30, auto_load=True, loadable=False)
class PaginationMiddleware(web.InternalMiddleware):
    async def handle(self, request: Request, params: dict[str, Any],
                     next_func: Callable[[Request, dict[str, Any]], Awaitable[Response]]):
        if 'page' in request.query or 'per_page' in request.query:
            try:
                page = int(request.query.get('page', "0"))
                per_page = int(request.query.get('per_page', "20"))
            except ValueError:
                raise BadRequestError('global.pagination.param_not_number')
            params['pagination'] = PaginationParams(page, per_page)
        return await next_func(request, params)


@middleware('blnt_query_order_by', priority=30, auto_load=True, loadable=False)
class OrderByMiddleware(web.InternalMiddleware):
    async def handle(self, request: Request, params: dict[str, Any],
                     next_func: Callable[[Request, dict[str, Any]], Awaitable[Response]]):
        if 'order_by' in request.query:
            order_by = []
            columns = request.query['order_by'].split(',')
            for column in columns:
                col_name, *col_args = column.split(':')
                order_way = col_args[0] if len(col_args) > 0 else 'asc'
                order_by.append(OrderByParams(col_name, order_way == 'asc'))
            params['order_by'] = order_by
        return await next_func(request, params)



import datetime
import json

from aiohttp.web_request import Request


async def deserialize(request: Request):
    content_type = request.content_type
    if content_type == 'application/json':
        return await request.json()
    if content_type == 'multipart/form-data':
        return await request.post()
    return {}


class Serializer:
    def __init__(self, mime, priority):
        self.mime = mime
        self.priority = priority

    def serialize(self, response):
        pass


class JSONSerializer(Serializer):
    def __init__(self):
        super().__init__('application/json', 1)

    @staticmethod
    def _json_type_converter(o):
        if isinstance(o, (datetime.date, datetime.datetime)):
            return o.isoformat()
        return o

    def serialize(self, response):
        return json.dumps(response, default=JSONSerializer._json_type_converter)


class Serializers:
    def __init__(self, functions=None):
        self.priorities = []
        self.serializers = {}
        if functions is not None:
            for serializer in functions:
                self.add(serializer)

    def add(self, serializer):
        self.serializers[serializer.mime] = serializer
        self.priorities = sorted(self.priorities + [serializer], key=lambda s: s.priority)

    def get(self, mime):
        return self.serializers.get(mime, None)

    @property
    def default(self):
        return self.priorities[0]


serializers = Serializers([
    JSONSerializer(),
])


def serialize(response, mime):
    serializer = serializers.get(mime) or serializers.default
    return serializer.serialize(response), serializer.mime



from bolinette.console import Console, console
from bolinette.extensions import BolinetteExtension, Extensions
import bolinette.defaults
import bolinette.init
import bolinette.commands
from bolinette.bolinette import Bolinette



import sys
from typing import Any

from bolinette import blnt, Console
from bolinette.utils import files, paths
from bolinette.decorators import command


def _create_file(console: Console, *,
                 source_dir: str, source_template: str, dest_dir: str, dest_file: str,
                 params: dict[str, Any]):
    if not paths.exists(dest_dir):
        console.error(f'Folder {dest_dir} not found.\n'
                      'Make sure you have set the module in the manifest or you are '
                      'using the CLI at project root directory.')
        sys.exit(1)
    dest = paths.join(dest_dir, dest_file)
    if paths.exists(dest):
        console.error(f'File {dest} already exist.')
        sys.exit(1)
    rendered = files.render_template(source_dir, source_template, params)
    files.write(dest, rendered)


def _update_init(path: str, module: str, folder: str, name: str, class_name: str):
    file_path = paths.join(path, '__init__.py')
    files.write(file_path, f'from {module}.{folder}.{name} import {class_name}\n', mode='a+')


def _create_model(context: 'blnt.BolinetteContext', console: Console, name: str):
    module = context.manifest['module']
    dest_dir = context.root_path(module, 'models')
    class_name = name[0].upper() + name[1:]
    _create_file(console,
                 source_dir=context.internal_files_path('cli', 'templates'),
                 source_template='model.py.jinja2',
                 dest_dir=dest_dir,
                 dest_file=f'{name}.py',
                 params={'name': name, 'class': class_name})
    _update_init(dest_dir, module, 'models', name, class_name)


def _create_service(context: 'blnt.BolinetteContext', console: Console, name: str):
    module = context.manifest['module']
    dest_dir = context.root_path(module, 'services')
    class_name = name[0].upper() + name[1:]
    _create_file(console,
                 source_dir=context.internal_files_path('cli', 'templates'),
                 source_template='service.py.jinja2',
                 dest_dir=dest_dir,
                 dest_file=f'{name}.py',
                 params={'name': name, 'class': class_name})
    _update_init(dest_dir, module, 'services', name, f'{class_name}Service')


def _create_controller(context: 'blnt.BolinetteContext', console: Console, name: str):
    module = context.manifest['module']
    dest_dir = context.root_path(module, 'controllers')
    class_name = name[0].upper() + name[1:]
    _create_file(console,
                 source_dir=context.internal_files_path('cli', 'templates'),
                 source_template='controller.py.jinja2',
                 dest_dir=dest_dir,
                 dest_file=f'{name}.py',
                 params={'name': name, 'class': class_name, 'module': module})
    _update_init(dest_dir, module, 'controllers', name, f'{class_name}Controller')


def _create_middleware(context: 'blnt.BolinetteContext', console: Console, name: str):
    module = context.manifest['module']
    dest_dir = context.root_path(module, 'middlewares')
    class_name = name[0].upper() + name[1:]
    _create_file(console,
                 source_dir=context.internal_files_path('cli', 'templates'),
                 source_template='middleware.py.jinja2',
                 dest_dir=dest_dir,
                 dest_file=f'{name}.py',
                 params={'name': name, 'class': class_name})
    _update_init(dest_dir, module, 'middlewares', name, f'{class_name}Middleware')


def _create_mixin(context: 'blnt.BolinetteContext', console: Console, name: str):
    module = context.manifest['module']
    dest_dir = context.root_path(module, 'mixins')
    class_name = name[0].upper() + name[1:]
    _create_file(console,
                 source_dir=context.internal_files_path('cli', 'templates'),
                 source_template='mixin.py.jinja2',
                 dest_dir=dest_dir,
                 dest_file=f'{name}.py',
                 params={'name': name, 'class': class_name})
    _update_init(dest_dir, module, 'mixins', name, f'{class_name}')


@command('new model', 'Create a new model file')
@command.argument('argument', 'name', summary='The new model\'s name')
@command.argument('flag', 'service', flag='s', summary='Create an associated service')
@command.argument('flag', 'controller', flag='c', summary='Create an associated controller')
async def create_model(context: 'blnt.BolinetteContext', name: str, service: bool, controller: bool):
    console = Console()
    _create_model(context, console, name)
    if service:
        _create_service(context, console, name)
    if controller:
        _create_controller(context, console, name)


@command('new service', 'Create a new service file')
@command.argument('argument', 'name', summary='The new service\'s name')
@command.argument('flag', 'controller', flag='c', summary='Create an associated controller')
async def create_service(context: 'blnt.BolinetteContext', name: str, controller: bool):
    console = Console()
    _create_service(context, console, name)
    if controller:
        _create_controller(context, console, name)


@command('new controller', 'Create a new controller file')
@command.argument('argument', 'name', summary='The new controller\'s name')
async def create_controller(context: 'blnt.BolinetteContext', name: str):
    console = Console()
    _create_controller(context, console, name)


@command('new middleware', 'Create a new middleware file')
@command.argument('argument', 'name', summary='The new middleware\'s name')
async def create_middleware(context: 'blnt.BolinetteContext', name: str):
    console = Console()
    _create_middleware(context, console, name)


@command('new mixin', 'Create a new mixin file')
@command.argument('argument', 'name', summary='The new mixin\'s name')
async def create_mixin(context: 'blnt.BolinetteContext', name: str):
    console = Console()
    _create_mixin(context, console, name)



from datetime import datetime

from bolinette import types, core, mapping
from bolinette.decorators import mixin


@mixin('historized')
class Historized(core.Mixin):
    def columns(self):
        return {
            'created_on': types.defs.Column(types.db.Date, nullable=False),
            'updated_on': types.defs.Column(types.db.Date, nullable=False),
            'created_by_id': types.defs.Column(
                types.db.Integer, reference=types.defs.Reference('user', 'id'), nullable=False),
            'updated_by_id': types.defs.Column(
                types.db.Integer, reference=types.defs.Reference('user', 'id'), nullable=False)
        }

    def relationships(self, model):
        return {
            'created_by': types.defs.Relationship('user', foreign_key=model.created_by_id, lazy=False),
            'updated_by': types.defs.Relationship('user', foreign_key=model.updated_by_id, lazy=False)
        }

    def response(self, model):
        return [
            mapping.Column(model.created_on),
            mapping.Column(model.updated_on),
            mapping.Reference(model.created_by),
            mapping.Reference(model.updated_by),
        ]

    @mixin.service_method
    async def create(self, values, current_user):
        if not current_user:
            return
        now = datetime.utcnow()
        values['created_on'] = now
        values['created_by_id'] = current_user.id
        values['updated_on'] = now
        values['updated_by_id'] = current_user.id

    @mixin.service_method
    async def update(self, entity, values, current_user):
        if not current_user:
            return
        now = datetime.utcnow()
        values['created_on'] = entity.created_on
        values['created_by_id'] = entity.created_by_id
        values['updated_on'] = now
        values['updated_by_id'] = current_user.id

    @mixin.service_method
    async def patch(self, values, current_user):
        if not current_user:
            return
        now = datetime.utcnow()
        values['updated_on'] = now
        values['updated_by_id'] = current_user.id



from bolinette.utils import files, paths
from setuptools import setup, find_packages

from bolinette import blnt


context = blnt.BolinetteContext(paths.join(paths.dirname(__file__), 'bolinette'))


def project_packages(module):
    return [m for m in find_packages() if m.startswith(module)]


setup(
    name='Bolinette',
    packages=project_packages('bolinette'),
    include_package_data=True,
    version=context.manifest['version'],
    license='MIT',
    description='An asynchronous Python web framework',
    long_description=files.read_file(context.root_path('README.md')),
    long_description_content_type='text/markdown',
    author='Pierre Chat',
    author_email='pierrechat@outlook.com',
    url='https://github.com/TheCaptainCat/bolinette',
    keywords=['Bolinette', 'Web', 'Framework', 'Async'],
    install_requires=files.read_requirements(context.root_path()),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.9',
    ],
    setup_requires=[
        'wheel'
    ],
    entry_points = {
        'console_scripts': ['blnt=blnt:main'],
    }
)



from collections.abc import Iterator as _Iterator


class APIError(Exception):
    def __init__(self, message: str, params: list[str] = None):
        super().__init__(type(self).__name__)
        self._message = message
        self._params = params or []

    def __str__(self):
        return ':'.join([self._message] + self._params)

    def __repr__(self):
        return f'<APIError {str(self)}>'

    def __getitem__(self, index: int):
        return self._params[index]

    def __setitem__(self, index: int, value: str):
        self._params = self._params[:index - 1] + [value] + self._params[index+1:]

    @property
    def message(self):
        return self._message


class APIErrors(Exception):
    def __init__(self):
        self.errors = []

    def append(self, error: APIError):
        self.errors.append(error)

    def __bool__(self):
        return len(self.errors) > 0

    def __iter__(self) -> _Iterator[APIError]:
        return iter(self.errors)

    def __str__(self):
        return str(self.errors)

    def __repr__(self):
        return f'<APIErrors [{",".join([repr(err) for err in self.errors])}]>'


class InternalError(APIError):
    pass


class NotFoundError(APIError):
    pass


class ConflictError(APIError):
    pass


class BadRequestError(APIError):
    pass


class UnprocessableEntityError(APIError):
    pass


class ForbiddenError(APIError):
    pass


class UnauthorizedError(APIError):
    pass


class EntityNotFoundError(NotFoundError):
    def __init__(self, model: str, key: str, value: str):
        super().__init__(f'entity.not_found', [model, key, value])


class ParamMissingError(UnprocessableEntityError):
    def __init__(self, key: str):
        super().__init__(f'param.required', [key])


class ParamNonNullableError(UnprocessableEntityError):
    def __init__(self, key: str):
        super().__init__(f'param.non_nullable', [key])


class BadParamFormatError(UnprocessableEntityError):
    def __init__(self, key: str, p_type: str):
        super().__init__(f'param.bad_format', [key, p_type])


class ParamConflictError(ConflictError):
    def __init__(self, key, value):
        super().__init__(f'param.conflict', [key, value])


class InitError(Exception):
    def __init__(self, message: str = None, *, inner: Exception = None):
        self.message = message
        self.inner = inner

    def __str__(self):
        if self.message is None:
            return str(self.inner)
        return self.message



import sqlalchemy
from sqlalchemy import orm as sqlalchemy_orm

from bolinette import blnt, core, web, Extensions
from bolinette.decorators import init_func
from bolinette.exceptions import InitError
from bolinette.utils import InitProxy


@init_func(extension=Extensions.MODELS)
async def init_model_classes(context: blnt.BolinetteContext):
    models = {}
    proxies = {}

    # Instantiate models & process columns
    for model_name, model_cls in blnt.cache.models.items():
        db_key = model_cls.__blnt__.database
        if db_key not in context.db:
            raise InitError(f'Undefined "{db_key}" database for model "{model_name}"')
        model = model_cls(context.db[db_key])
        # Instantiate columns
        for col_name, proxy in model.__props__.get_proxies(core.models.Column):
            col = proxy.instantiate(name=col_name, model=model)
            proxies[proxy] = col
            setattr(model, col_name, col)
        # Add mixin columns
        for mixin_name in model.__blnt__.mixins:
            if mixin_name not in blnt.cache.mixins:
                raise InitError(f'Model "{model_name}": mixin "{mixin_name}" is not defined')
            mixin = blnt.cache.mixins[mixin_name]()
            for col_name, proxy in mixin.columns().items():
                col = proxy.instantiate(name=col_name, model=model)
                setattr(model, col_name, col)
            model.__props__.mixins[mixin_name] = mixin
        models[model_name] = model

    for model_name, model in models.items():
        # Process auto generated primary keys
        primary = [c for _, c in model.__props__.get_columns() if c.primary_key]
        if not primary:
            model.__props__.primary = None
        else:
            if len(primary) == 1 and primary[0].auto_increment is None:
                primary[0].auto_increment = True
            model.__props__.primary = primary
        for column in primary:
            column.nullable = False
        # Find entity key
        entity_key = [c for _, c in model.__props__.get_columns() if c.entity_key]
        if not entity_key:
            model.__props__.entity_key = model.__props__.primary
        else:
            model.__props__.entity_key = entity_key
        if model.__props__.entity_key is None:
            raise InitError(f'No entity key defined for "{model_name}" model. '
                            'Mark one column or more as entity_key=True.')

    # Process relationships
    for model_name, model in models.items():
        # Instantiate relationships
        for rel_name, proxy in model.__props__.get_proxies(core.models.Relationship):
            rel = proxy.instantiate(name=rel_name, model=model, models=models)
            proxies[proxy] = rel
            setattr(model, rel_name, rel)
        # Add mixin relationships
        for _, mixin in model.__props__.mixins.items():
            for rel_name, proxy in mixin.relationships(model).items():
                rel = proxy.instantiate(name=rel_name, model=model, models=models)
                setattr(model, rel_name, rel)

    # Process references
    for model_name, model in models.items():
        # Instantiate references
        for col_name, col in model.__props__.get_columns():
            if isinstance(col.reference, InitProxy) and col.reference.of_type(core.models.Reference):
                col.reference = col.reference.instantiate(model=model, column=col, models=models)
        # Process back references
        added_back_refs: dict[core.Model, list[core.models.ColumnList]] = {}
        for rel_name, rel in model.__props__.get_relationships():
            # Instantiate back references
            if isinstance(rel.backref, InitProxy) and rel.backref.of_type(core.models.Backref):
                rel.backref = rel.backref.instantiate(model=model, relationship=rel)
                if rel.backref.key not in added_back_refs:
                    added_back_refs[rel.target_model] = []
                added_back_refs[rel.target_model].append(
                    core.models.ColumnList(rel.backref.key, rel.target_model, model))
            # Link foreign keys
            if isinstance(rel.foreign_key, InitProxy) and rel.foreign_key.of_type(core.models.Column):
                rel.foreign_key = proxies[rel.foreign_key]
            # Link remote side foreign keys for same-model-references
            if isinstance(rel.remote_side, InitProxy) and rel.remote_side.of_type(core.models.Column):
                rel.remote_side = proxies[rel.remote_side]
        # Add back refs to the target model
        for target_model, back_refs in added_back_refs.items():
            for back_ref in back_refs:
                setattr(target_model, back_ref.name, back_ref)

    for model_name, model in models.items():
        context.add_model(model_name, model)


@init_func(extension=Extensions.MODELS)
async def init_relational_models(context: blnt.BolinetteContext):
    models = {}
    for model_name, model in context.models:
        if model.__props__.database.relational:
            models[model_name] = model
    orm_tables = {}
    orm_cols = {}
    for model_name, model in models.items():
        orm_cols[model_name] = {}
        for att_name, attribute in model.__props__.get_columns():
            attribute.name = att_name
            ref = None
            if attribute.reference:
                ref = sqlalchemy.ForeignKey(attribute.reference.target_path)
            orm_cols[model_name][att_name] = sqlalchemy.Column(
                att_name, attribute.type.sqlalchemy_type, ref,
                default=attribute.default, index=attribute.entity_key,
                primary_key=attribute.primary_key, nullable=attribute.nullable,
                unique=attribute.unique, autoincrement=attribute.auto_increment)
        orm_tables[model_name] = sqlalchemy.Table(model_name,
                                                  model.__props__.database.base.metadata,
                                                  *(orm_cols[model_name].values()))

    for model_name, model in models.items():
        orm_defs = {}
        for att_name, attribute in model.__props__.get_relationships():
            kwargs = {}
            attribute.name = att_name
            if attribute.backref:
                kwargs['backref'] = sqlalchemy_orm.backref(attribute.backref.key, lazy=attribute.backref.lazy)
            if attribute.foreign_key:
                kwargs['foreign_keys'] = orm_cols[model_name][attribute.foreign_key.name]
            if attribute.remote_side:
                kwargs['remote_side'] = orm_cols[model_name][attribute.remote_side.name]
            if attribute.secondary:
                kwargs['secondary'] = orm_tables[attribute.secondary.__blnt__.name]
            orm_defs[att_name] = sqlalchemy_orm.relationship(attribute.target_model_name, lazy=attribute.lazy, **kwargs)

        orm_defs['__table__'] = orm_tables[model_name]
        orm_model = type(model_name, (model.__props__.database.base,), orm_defs)

        for att_name, attribute in model.__props__.get_properties():
            setattr(orm_model, att_name, property(attribute.function))

        context.add_table(model_name, orm_model)


@init_func(extension=Extensions.MODELS)
async def init_databases(context: blnt.BolinetteContext):
    await context.db.create_all()


@init_func(extension=Extensions.MODELS)
async def init_repositories(context: blnt.BolinetteContext):
    for model_name, model in context.models:
        repo = core.Repository(model_name, model, context)
        model.__repo__ = repo
        context.add_repo(model_name, repo)


@init_func(extension=Extensions.MODELS)
async def init_mappings(context: blnt.BolinetteContext):
    for model_name, model in context.models:
        context.mapper.register(model_name, model)


@init_func(extension=Extensions.MODELS)
async def init_services(context: blnt.BolinetteContext):
    for service_name, service_cls in blnt.cache.services.items():
        context.add_service(service_name, service_cls(context))


@init_func(extension=Extensions.WEB)
async def init_controllers(context: blnt.BolinetteContext):
    def _init_route(_ctrl: web.Controller, _proxy: InitProxy):
        _route = _proxy.instantiate(controller=_ctrl)
        if _route.inner_route is not None:
            _route.inner_route, _route.func = _init_route(_ctrl, _route.inner_route)
        return _route, _route.func

    for controller_name, controller_cls in blnt.cache.controllers.items():
        controller = controller_cls(context)
        for route_name, proxy in controller.__props__.get_proxies(web.ControllerRoute):
            route, _ = _init_route(controller, proxy)
            setattr(controller, route_name, route)
        for _, route in controller.__props__.get_routes():
            route.controller = controller
            route.setup()
        for route in controller.default_routes():
            route.controller = controller
            route.setup()
        context.add_controller(controller_name, controller)


@init_func(extension=Extensions.SOCKETS)
async def init_topics(context: blnt.BolinetteContext):
    for topic_name, topic_cls in blnt.cache.topics.items():
        topic = topic_cls(context)
        context.sockets.add_topic(topic_name, topic)
        for _, channel in topic.__props__.get_channels():
            context.sockets.add_channel(topic_name, channel)



from bolinette.blnt.logging import Logger
from bolinette.blnt.props import Properties
from bolinette.blnt.cache import cache, InitFunc
from bolinette.blnt.environment import init, Environment
from bolinette.blnt.jwt import JWT
from bolinette.blnt.validation import Validator
from bolinette.blnt.context import BolinetteContext
from bolinette.blnt.transaction import Transaction



import re

from pymongo import MongoClient

from bolinette.blnt.database.engines import DatabaseEngine

_COLLECTION_REGEX = re.compile(r'^([^/]+://[^/]+/?)(.*)$')


class CollectionDatabase(DatabaseEngine):
    def __init__(self, uri):
        super().__init__(relational=False)
        db = 'bolinette'
        if match := _COLLECTION_REGEX.match(uri):
            uri = match.group(1)
            db = match.group(2) or db
        self.client = MongoClient(uri)
        self.db = self.client[db]

    async def open_transaction(self):
        pass

    async def close_transaction(self):
        pass

    async def rollback_transaction(self):
        pass

    async def create_all(self):
        pass

    async def drop_all(self):
        for collection in self.db.list_collection_names():
            self.db[collection].delete_many({})



from typing import Any

from bolinette import blnt, core
from bolinette.blnt.objects import Pagination, PaginationParams, OrderByParams
from bolinette.blnt.database.queries import BaseQuery, RelationalQueryBuilder, CollectionQueryBuilder
from bolinette.exceptions import ParamConflictError, APIErrors, ParamNonNullableError
from bolinette.utils.functions import setattr_, getattr_, async_invoke


class Repository:
    def __init__(self, name: str, model: 'core.Model', context: 'blnt.BolinetteContext'):
        self.name = name
        self.model = model
        self.context = context
        self._query_builder = self._get_query_builder(model, context)

    @staticmethod
    def _get_query_builder(model: 'core.Model', context: 'blnt.BolinetteContext'):
        db = context.db[model.__blnt__.database]
        if db.relational:
            return RelationalQueryBuilder(model, context)
        return CollectionQueryBuilder(model, context)

    def __repr__(self):
        return f'<Repository {self.name}>'

    def query(self):
        return self._query_builder.query()

    async def get_all(self, pagination: PaginationParams = None, order_by: list[OrderByParams] = None):
        if order_by is None:
            order_by = []
        query = self.query()
        if len(order_by) > 0:
            for order in order_by:
                query = query.order_by(order.column, desc=not order.ascending)
        if pagination is not None:
            return await self._paginate(query, pagination)
        return await query.all()

    async def get(self, identifier):
        return await self.query().get_by_id(identifier)

    async def get_by(self, key: str, value):
        return await self.query().filter_by(**{key: value}).all()

    async def get_first_by(self, key: str, value):
        return await self.query().filter_by(**{key: value}).first()

    async def create(self, values: dict[str, Any], **kwargs):
        await self._call_mixin_methods('create', self, values=values, **kwargs)
        filtered = await self._validate_model(values, kwargs)
        entity = await self._query_builder.insert_entity(filtered)
        return entity

    async def update(self, entity, values: dict[str, Any], **kwargs):
        await self._call_mixin_methods('update', self, entity=entity, values=values, **kwargs)
        await self._map_model(entity, values)
        return await self._query_builder.update_entity(entity)

    async def patch(self, entity, values: dict[str, Any], **kwargs):
        await self._call_mixin_methods('patch', self, entity=entity, values=values, **kwargs)
        await self._map_model(entity, values, patch=True)
        return await self._query_builder.update_entity(entity)

    async def delete(self, entity, **kwargs):
        await self._call_mixin_methods('delete', self, entity=entity, **kwargs)
        return await self._query_builder.delete_entity(entity)

    @staticmethod
    async def _paginate(query: BaseQuery, pagination: PaginationParams):
        page = pagination.page
        per_page = pagination.per_page
        total = await query.count()
        items = await query.offset(page * per_page).limit(per_page).all()
        return Pagination(items, page, per_page, total)

    async def _validate_model(self, values: dict[str, Any], repo_args: dict[str, Any]):
        api_errors = APIErrors()
        ignore_keys: list[str] = []
        for _, relationship in self.model.__props__.get_relationships():
            await self._validate_linked_model(relationship, values, repo_args, ignore_keys)
        for _, column in self.model.__props__.get_columns():
            key = column.name
            if column.auto_increment or key in ignore_keys:
                continue
            if key in values:
                value = values.get(key)
                if value is None and not column.nullable:
                    api_errors.append(ParamNonNullableError(key))
                if column.unique:
                    if await self.get_first_by(column.name, value) is not None:
                        api_errors.append(ParamConflictError(key, value))
            elif not column.nullable:
                api_errors.append(ParamNonNullableError(key))
        if api_errors:
            raise api_errors
        return values

    @staticmethod
    async def _validate_linked_model(relationship: 'core.models.Relationship', values: dict[str, Any],
                                     repo_args: dict[str, Any], ignore_keys: list[str]):
        if relationship.foreign_key is None:
            return
        foreign_key = relationship.foreign_key.name
        if values.get(foreign_key) is not None:
            return
        key = relationship.name
        if key in values:
            ignore_keys.append(foreign_key)
            linked = values[key]
            if isinstance(linked, dict):
                values[key] = await relationship.target_model.__repo__.create(values[key], **repo_args)

    async def _map_model(self, entity, values: dict[str, Any], patch=False):
        api_errors = APIErrors()
        for _, column in self.model.__props__.get_columns():
            key = column.name
            if column.primary_key or (key not in values and patch):
                continue
            original = getattr_(entity, key, None)
            new = values.get(key, None)
            if original == new:
                continue
            if new is None and not column.nullable:
                api_errors.append(ParamNonNullableError(key))
            if column.unique and new is not None:
                if await self.get_first_by(column.name, new) is not None:
                    api_errors.append(ParamConflictError(key, new))
            setattr_(entity, key, new)
        if api_errors:
            raise api_errors

    async def _call_mixin_methods(self, method_name: str, *args, **kwargs):
        for _, mixin in self.model.__props__.mixins.items():
            if method_name in mixin:
                await async_invoke(mixin[method_name].func, *args, **kwargs)
